"""
kzconfig.context
~~~~~

Kazoo config library.

"""
from urllib.parse import urlparse, unquote

import pyrkube

from . import meta, util
from .dns import DNS
from .couch import CouchDB
from .kazoo import Kazoo
from .sup import Sup
from .kubectl import KubeCtl
from .kube import Kube


class Context(metaclass=meta.Singleton):
    _configs = ('environment',)
    _secrets = ('couchdb', 'rabbitmq', 'master-account', 'dns.dnsimple')

    def __init__(self, domain='telephone.org'):
        self.domain = domain

    def _get_configmap(self, name):
        return self.kube.api.get('configmap', name).data

    def _get_secret(self, name):
        return self.kube.api.get('secret', name).data

    @meta.lazy_property
    def configs(self):
        try:
            return {cm: self._get_configmap(cm) for cm in self._configs}
        except AttributeError:
            pass

    @meta.lazy_property
    def secrets(self):
        try:
            return {s: self._get_secret(s) for s in self._secrets}
        except AttributeError:
            pass

    @meta.lazy_property
    def kube(self):
        return Kube(self)

    @meta.lazy_property
    def couchdb(self):
        return CouchDB(self)

    @meta.lazy_property
    def kazoo(self):
        return Kazoo(self)

    @meta.lazy_property
    def dns(self):
        return DNS(self)

    @meta.lazy_property
    def sup(self):
        return Sup(self)

    @meta.lazy_property
    def kubectl(self):
        return KubeCtl(self)



"""
kzconfig.util
~~~~~

Kazoo config library.

"""

import base64
import json
from os.path import join

import dns.resolver


def quote(content):
    return '"' + content + '"'


def b64encode(data):
    if isinstance(data, str):
        data = data.encode()
    return base64.b64encode(data).decode()


def b64decode(data):
    if isinstance(data, str):
        data = data.encode()
    return base64.b64encode(data).decode()


def join_url(scheme, user, password, host):
    return '{}://{}:{}@{}'.format(scheme, user, password, host)


def json_dumps(obj, safe=True):
    seperators = (',', ':') if safe else None
    return json.dumps(obj, separators=seperators)


def addrs_for(*domains):
    addrs = []
    for domain in domains:
        ans = [a.address for a in dns.resolver.query(domain, 'A')]
        addrs.extend(ans)
    return list(set(addrs))


def to_str(items, path='/', flag='data', delim=' '):
    return delim.join('--{} {}/{}'.format(flag, path, item) for item in items)


def read_config(path, base='.'):
    with open(join(base, path)) as fd:
        return fd.read().strip()



from invoke import task


@task
def develop(ctx):
    ctx.run("python3 setup.py develop")


@task
def build(ctx):
    ctx.run("python3 setup.py sdist bdist_wheel")


@task
def upload(ctx):
    ctx.run("twine upload dist/*")


@task
def clean(ctx):
    ctx.run("rm -rf build dist *.egg-info **/.pyc **/__pycache__")


@task(build, upload, clean)
def publish(ctx):
    pass



import os
from os.path import join
from pathlib import Path
import sys
import hashlib

import requests
import click


KUBECTL_URL_TEMPLATE = (
    'https://storage.googleapis.com/kubernetes-release/release/'
    'v{version}/bin/{os}/amd64/kubectl'
)
KUBECTL_VERSION = os.getenv('KUBECTL_VERSION', '1.7.8')
KUBECTL_SHA256 = dict(
    linux=os.getenv(
        'KUBECTL_SHA256',
        'c4fd350f9fac76121dda479c1ceba2d339b19f8806aa54207eff55c9c6896724'
    ),
    darwin=os.getenv(
        'KUBECTL_SHA256',
        '9dbfc0337ab7aabf4b719eeffba77b42afaf361a3cda3343dccf6920e25f6085'
    )
)
KUBECTL_DOWNLOAD_DIR = Path(os.getenv('KUBECTL_DOWNLOAD_DIR', '/usr/local/bin'))
KUBECTL_LINK_PATH = KUBECTL_DOWNLOAD_DIR / 'kubectl'
KUBE_SERVICEACCOUNT_PATH = Path('/var/run/secrets/kubernetes.io/serviceaccount')
KUBECTL_CONFIG_PATH = Path(os.path.expanduser(join('~', '.kube', 'config2')))
KUBECTL_CONFIG = """apiVersion: v1
clusters:
- cluster:
    certificate-authority: /var/run/secrets/kubernetes.io/serviceaccount/ca.crt
    server: https://kubernetes.default
  name: local
contexts:
- context:
    cluster: local
    namespace: default
    user: admin
  name: local
current-context: local
kind: Config
preferences: {}
users:
- name: admin
  user:
    token: default"""


@click.command()
def download():
    path = KUBECTL_DOWNLOAD_DIR / 'kubectl-{}'.format(KUBECTL_VERSION)
    if not path.exists():
        try:
            path.parent.mkdir(parents=True)
        except FileExistsError:
            pass

        platform = sys.platform
        url = KUBECTL_URL_TEMPLATE.format(version=KUBECTL_VERSION, os=platform)
        r = requests.get(url, stream=True)
        r.raise_for_status()

        filename = path.with_name('{}.download'.format(path.name))
        sha256 = hashlib.sha256()
        with filename.open('wb') as fd:
            for chunk in r.iter_content(chunk_size=4096):
                if chunk:
                    fd.write(chunk)
                    sha256.update(chunk)

        if sha256.hexdigest() != KUBECTL_SHA256[platform]:
            raise RuntimeError('Checksum mismatch for kubectl')

        filename.chmod(0o755)
        filename.rename(path)
        if KUBECTL_LINK_PATH.exists():
            os.remove(KUBECTL_LINK_PATH)
        KUBECTL_LINK_PATH.symlink_to(path)
        write_kubeconfig()
    return str(path)


def write_kubeconfig():
    if not KUBECTL_CONFIG_PATH.exists():
        with KUBECTL_CONFIG_PATH.open('w') as fd:
            fd.write(KUBECTL_CONFIG)



"""
kzconfig.sup
~~~~~

Kazoo config library.


"""

import subprocess
from decimal import Decimal

from . import util


class SupCommandBase:
    def __init__(self, parent):
        self.parent = parent
        self.sup = parent.sup


class KappsMaint(SupCommandBase):
    module = 'kapps_maintenance'

    def refresh(self, database=''):
        return self.sup(self.module, 'refresh', database)


class KappsAcctConfig(SupCommandBase):
    module = 'kapps_account_config'

    def get(self, acct, doc, key, default='undefined'):
        if not isinstance(default, str):
            default = str(default)
        return self.sup(self.module, 'get', acct, doc, key, default)

    def set(self, acct, doc, key, value):
        if isinstance(value, bytes):
            value = value.decode()
        return self.sup(self.module, 'set', acct, doc, key, value)

    def flush(self, acct, doc=None, strategy=None):
        args = [acct, doc]
        if strategy:
            args.append(strategy)
        return self.sup(self.module, 'flush', *args)


class KappsConfig(SupCommandBase):
    module = 'kapps_config'

    def get(self, doc, key, default='undefined', node=None):
        if not isinstance(default, str):
            default = str(default)

        args = [doc, key, default]
        if node:
            args.append(node)
        return self.sup(self.module, 'get', *args)

    def set(self, doc, key, value, node=None):
        if isinstance(value, bytes):
            value = value.decode()

        if isinstance(value, bool):
            func = 'set_boolean'
            value = str(value).lower()
        elif isinstance(value, int):
            func = 'set_integer'
        elif isinstance(value, (float, Decimal)):
            func = 'set_float'
            value = str(value)
        else:
            func = 'set'

        args = [doc, key, value]
        if node:
            args.append(node)
        return self.sup(self.module, func, *args)

    def set_json(self, doc, key, value):
        if not isinstance(value, str):
            value = util.json_dumps(value)

        return self.sup(self.module, 'set_json', doc, key, value)

    def set_default(self, doc, key, value):
        return self.sup(self.module, 'set_default', doc, key, value)

    def flush(self, doc=None, key=None, node=None):
        args = []
        for arg in (doc, key, node):
            if arg:
                args.append(arg)

        return self.sup(self.module, 'flush', *args)


class KazooNodes(SupCommandBase):
    module = 'kz_nodes'
    def status(self):
        return self.sup(self.module, 'status')


class Sup:
    def __init__(self, context):
        self.context = context
        self.kapps_maintenance = KappsMaint(self)
        self.kapps_account_config = KappsAcctConfig(self)
        self.kapps_config = KappsConfig(self)
        self.kz_nodes = KazooNodes(self)

    def sup(self, module, function, *args):
        args = list(args)
        try:
            pod = self.context.kube.api.get_first('pod', selector=dict(app='kazoo'))
        except TypeError:
            raise RuntimeError('No kazoo app pod found')

        for idx, arg in enumerate(args):
            if isinstance(arg, (int, bool)):
                args[idx] = str(arg)

        args_str = ' '.join([module, function, *args])
        cmd = 'kubectl exec {} -- sup {}'.format(pod.name, args_str)
        print(cmd)
        return subprocess.getoutput(cmd)

    def sup_api(self, module, function, *args):
        _, result = self.context.kazoo.sup(module, function, *args)
        return result



class Singleton(type):
    """Metaclass for singletons. Any instantiation of a Singleton class yields
    the exact same object, e.g.:

    >>> class MyClass(metaclass=Singleton):
            pass
    >>> a = MyClass()
    >>> b = MyClass()
    >>> a is b
    True
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args, **kwargs)
        else:
            cls._instances[cls].__init__(*args, **kwargs)
        return cls._instances[cls]

    @classmethod
    def __instancecheck__(mcs, instance):
        if instance.__class__ is mcs:
            return True
        else:
            return isinstance(instance.__class__, mcs)


# def singleton(cls):
#     obj = cls()
#     # Always return the same object
#     cls.__new__ = staticmethod(lambda cls: obj)
#     # Disable __init__
#     try:
#         del cls.__init__
#     except AttributeError:
#         pass
#     return cls

class lazy_property(object):
    '''
    meant to be used for lazy evaluation of an object attribute.
    property should represent non-mutable data, as it replaces itself.
    '''

    def __init__(self, fget):
        self.fget = fget
        self.func_name = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return None
        value = self.fget(obj)
        if value is not None or value == False:
            setattr(obj, self.func_name, value)
            return value
